;;; -*- Mode: Scheme; scheme48-package: stubber-code-generation -*-

;;;; Scheme48 Stubber
;;;; Code Generation

;;; This code is written by Taylor R. Campbell and placed in the Public
;;; Domain.  All warranties are disclaimed.

;;; This should be the only part of the stubber that knows anything
;;; about the grubby details of Scheme48's C interface.

;;;; Simple Value Bindings

(define (stub/binding scheme-name binding-name expression environment)
  (combine-stub
   (stub/binding/scheme scheme-name binding-name)
   (stub/binding/c binding-name expression environment)))

(define (stub/binding/scheme scheme-name binding-name)
  (stub/scheme-forms
   `(IMPORT-DEFINITION ,scheme-name ,binding-name)))

(define (stub/binding/c binding-name expression environment)
  (stub/c-initializations
   (let ((temporary (generate-c-temporary "value" environment)))
     (c-block
      (s48-declare-false temporary)
      (s48-declare-gc-protect 1)
      (s48-gc-protect (list temporary))
      ((process-expression expression environment) temporary)
      (s48-export-binding (output-escaped-string binding-name)
                          temporary)
      (s48-gc-unprotect)))))

;;;; Function Stubs

(define (stub/function scheme-name binding-name c-name
                       scheme-parameters
                       scheme-argument-expressions
                       scheme-output-processor
                       c-parameters c-declarations c-body)
  (combine-stub
   (stub/function/scheme scheme-name binding-name
                         scheme-parameters
                         scheme-argument-expressions
                         scheme-output-processor)
   (stub/function/c c-name binding-name
                    c-parameters
                    c-declarations
                    c-body)))

(define (stub/function/scheme scheme-name binding-name
                              parameter-names
                              argument-expressions
                              result-processor)
  (stub/scheme-forms
   (let ((binding-variable '%%%THE-BINDING%%%)) ;++ ick, want hygiene
     `(DEFINE ,scheme-name
        (LET ((,binding-variable
               (LOOKUP-IMPORTED-BINDING ,binding-name)))
          (LAMBDA ,parameter-names
            ,((lambda (call)
                (if result-processor
                    `(,result-processor ,call)
                    call))
              `(CALL-IMPORTED-BINDING ,binding-variable
                                      ,@argument-expressions))))))))

(define (stub/function/c c-name binding-name
                         parameters
                         declarations
                         body)
  (combine-stub
   (stub/c-declarations
    (c-stub-definition c-name parameters declarations body))
   (stub/c-initializations
    (c-stub-initialization binding-name c-name))))

(define (c-stub-initialization binding-name c-name)
  (s48-export-binding (output-escaped-string binding-name)
                      (c-call "s48_enter_pointer"
                              (list (c-cast "void *" c-name)))))

(define (c-stub-definition c-name parameters declarations body)
  (c-function (output-sequence "static " (s48-value-type))
      c-name
      (map (lambda (parameter)
             (output-sequence (s48-value-type)
                              " "
                              (stub-parameter-name parameter)))
           parameters)
    (c-stub-body (c-stub-result-variable c-name)
                 parameters
                 declarations
                 body)))

(define (c-stub-result-variable stub-name)             ;++ Foo.
  (output-sequence "__" stub-name "__result_s48"))

(define (c-stub-body result-variable parameters declarations body)
  (let ((protected-variables
         ;++ What about user-supplied protected variables?
         (cons result-variable (map stub-parameter-name parameters))))
    (output-sequence
     (c-stub-comment "Declarations")
     (output-list (map stub-parameter-declaration parameters))
     declarations
     (s48-declare-unspecific result-variable)
     (c-stub-comment "GC Protection")
     (s48-declare-gc-protect (length protected-variables))
     (s48-gc-protect protected-variables)
     (c-stub-comment "Conversions")
     (map stub-parameter-conversion parameters)
     (c-stub-comment "Body")
     (body result-variable)
     (c-stub-comment "Finish")
     (s48-gc-unprotect)
     (c-return result-variable))))

(define (c-stub-comment . elements)
  (output-sequence newline
                   (c-line (apply c-comment elements))
                   newline))

;;;; Enumeration Stubs

;++ I am *profusely* apologetic for the sorry state of this gruesome
;++ mess of a procedure.

(define (stub/enumeration scheme-name scheme-rtd
                          scheme-type-predicate
                          scheme-instance-vector
                          instance-vector-binding
                          scheme-name-accessor
                          scheme-index-accessor
                          scheme-available-predicate
                          available-predicate-binding
                          scheme-enumerand-names
                          enumerand-expressions
                          environment
                          c-type c-mapper c-unmapper)
  (let ((c-instance-vector
         (symbol->c-variable-name scheme-instance-vector environment))
        (c-available-predicate
         (symbol->c-stub-name scheme-available-predicate environment)))
    (let ((c-mapping-variable
           (output-sequence c-instance-vector "_mapping"))
          (c-unmapping-binding-variable
           (output-sequence c-instance-vector "_unmapping_binding")))
      (combine-stub
       (stub/scheme-forms
        `(DEFINE-ENUMERATED-TYPE ,scheme-name ,scheme-rtd
           ,scheme-type-predicate
           ,scheme-instance-vector
           ,scheme-name-accessor
           ,scheme-index-accessor
           ,scheme-enumerand-names)
        `(DEFINE-EXPORTED-BINDING ,instance-vector-binding
           ,scheme-instance-vector))
       (stub/c-declarations
        (c-enumeration-mapping c-type
                               enumerand-expressions
                               environment
                               c-mapping-variable)
        (s48-declare-false c-unmapping-binding-variable)
        (c-enumeration-mapper c-type c-mapper c-mapping-variable)
        (c-enumeration-unmapper c-type
                                c-unmapper
                                c-mapping-variable
                                c-unmapping-binding-variable
                                (length scheme-enumerand-names)))
       (stub/enumeration/predicate scheme-available-predicate
                                   available-predicate-binding
                                   c-available-predicate
                                   c-mapping-variable
                                   scheme-index-accessor)
       (stub/c-initializations
        (s48-gc-protect-global (list c-unmapping-binding-variable))
        (s48-export-binding
         (output-escaped-string instance-vector-binding)
         c-unmapping-binding-variable))))))

(define (stub/enumeration/predicate scheme-available-predicate
                                    available-predicate-binding
                                    c-available-predicate
                                    c-mapping-variable
                                    scheme-index-accessor)
  (let ((enumerand
         (if (eq? scheme-index-accessor 'ENUMERAND)
             '*ENUMERAND
             'ENUMERAND)))
    (stub/function
     scheme-available-predicate              ;Scheme name
     available-predicate-binding             ;Binding name
     c-available-predicate                   ;C stub name
     (list enumerand)                        ;Scheme parameter names
     `((,scheme-index-accessor ,enumerand))  ;Argument expressions
     #f                                      ;No output processor
     (list (stub-parameter/integral "int" "index"))  ;C stub parameters
     '()                                     ;No declarations
     (stub-result/boolean                    ;Stub result
      (c-field (c-array-ref c-mapping-variable "index")
               "available_p")))))

;;; These yield simple lines and not statements because the relevant
;;; macros expand to top-level function definitions.

(define (c-enumeration-mapper c-type c-mapper mapping-variable)
  (c-line
   (cpp-call "S48_ENUMERAND_MAPPER"
             (list c-mapper c-type mapping-variable))))

(define (c-enumeration-unmapper c-type
                                c-unmapper
                                mapping-variable
                                instance-vector-binding
                                count)
  (c-line
   (cpp-call "S48_ENUMERAND_UNMAPPER"
             (list c-unmapper
                   c-type
                   mapping-variable
                   instance-vector-binding
                   count))))

;++ This is ad-hoc and ugly.

(define (c-enumeration-mapping c-type
                               enumerand-expressions
                               environment
                               c-mapping-variable)
  (c-declaration
   "struct"
   (c-block (c-declare "char" "available_p")
            (c-declare c-type "value"))
   (output-sequence c-mapping-variable " [] ="     ;++ ick
                    (c-block
                     (map (lambda (enumerand)
                            (c-enumerand enumerand environment))
                          enumerand-expressions)))))

(define (c-enumerand enumerand environment)
  (define (winner expression)
    (c-line "{ 1, " (c-expression expression) " },"))
  (define (loser)
    (c-line "{ 0 },"))
  (match enumerand
    (('OR)
     (error "Empty enumerand option set."))
    ((OR (? string? expression)
         ('OR (? string? expression)))
     (cpp-ifdef expression
                (winner expression)
                (cpp-else)
                (loser)))
    (('OR (? string? expression) (? string? expressions) ___)
     (cpp-if (cpp-defined? expression)
             (winner expression)
             (fold-right (lambda (expression tail)
                           (output-sequence
                            (cpp-elif (cpp-defined? expression))
                            (winner expression)
                            tail))
                         (output-nothing)
                         expressions)
             (cpp-else)
             (loser)))
    (else
     (error "Malformed enumerand:" enumerand))))

;;;; Stub Parameters

(define-record-type* stub-parameter
  (make-stub-parameter name declaration conversion)
  ())

(define (stub-parameter/scheme name)
  (make-stub-parameter name
    (c-line (c-comment "No declaration for " name))
    (c-line (c-comment "No conversion for " name))))

(define (simple-stub-parameter name type extract)
  (let ((s48-name (output-sequence name "_s48")))
    (make-stub-parameter s48-name
      (c-declare type name)
      (c-statement
       (c-assign name
                 (c-cast type
                         (c-call extract (list s48-name))))))))

(define (stub-parameter/boolean name)
  (simple-stub-parameter name "int" "S48_EXTRACT_BOOLEAN"))

(define (stub-parameter/char name)
  (simple-stub-parameter name "char" "s48_extract_char"))

(define (stub-parameter/integral type name)
  (simple-stub-parameter name type "s48_extract_integer"))

(define (stub-parameter/floating-point type name)
  (simple-stub-parameter name type "s48_extract_double"))

(define (stub-parameter/shared-string name)
  (simple-stub-parameter name "char *" "s48_extract_string"))

(define (stub-parameter/copied-string name)
  (simple-stub-parameter name "char *" "s48_extract_and_copy_string"))

(define (stub-parameter/byte-pointer name)
  (simple-stub-parameter name "char *" "s48_extract_byte_vector"))

;;; An `alien' is a C value stored in a Scheme byte vector.  An `alien
;;; pointer' is not an alien whose value is a pointer, but a pointer
;;; inside the byte vector.  This may be a bad idea.

(define (alien-stub-parameter type name extract)
  (let ((s48-name (output-sequence name "_s48")))
    (make-stub-parameter s48-name
      (c-declare type name)
      (c-statement
       ;; Use CPP-CALL here so that we avoid extraneous parentheses.
       ;; This may be significant around the type.  I am not sure, but
       ;; it doesn't hurt to be safe.
       (c-assign name (cpp-call extract (list s48-name type)))))))

(define (stub-parameter/alien-pointer type name)
  (alien-stub-parameter type name "S48_EXTRACT_VALUE_POINTER"))

(define (stub-parameter/alien type name)
  (alien-stub-parameter type name "S48_EXTRACT_VALUE"))

(define (byte-vector-stub-parameter type bv-name length-name body)
  (let ((s48-name (output-sequence bv-name "_s48")))
    (make-stub-parameter s48-name
      (output-sequence (c-declare type bv-name)
                       (c-declare "size_t" length-name))
      (body s48-name))))

(define (stub-parameter/shared-byte-vector type bv-name length-name)
  (byte-vector-stub-parameter
   type bv-name length-name
   (lambda (s48-name)
     (output-sequence
      (c-statement
       (c-assign bv-name
                 (c-cast type
                         (c-call "s48_extract_byte_vector"
                                 (list s48-name)))))
      (c-statement
       (c-assign length-name
                 (c-call "S48_BYTE_VECTOR_LENGTH"
                         (list s48-name))))))))

(define (stub-parameter/copied-byte-vector type bv-name length-name)
  (byte-vector-stub-parameter
   type bv-name length-name
   (lambda (s48-name)
     (c-statement
      (c-assign
       bv-name
       (c-cast type
               (c-call "s48_extract_and_copy_byte_vector"
                       (list s48-name
                             (c-address-of length-name)))))))))

(define (stub-parameter/channel->fd name)
  (let ((s48-channel-name (output-sequence name "_channel_s48")))
    (make-stub-parameter s48-channel-name
      (c-declare "int" name)
      (c-statement
       (c-assign name
                 (c-call "S48_UNSAFE_EXTRACT_FIXNUM"
                         (list (c-call "S48_UNSAFE_CHANNEL_OS_INDEX"
                                       (list s48-channel-name)))))))))

(define (stub-parameter/false->null name parameter)
  (let ((s48-name (stub-parameter-name parameter)))
    (make-stub-parameter s48-name
      (stub-parameter-declaration parameter)
      (c-if (c-call "S48_FALSE_P" (list s48-name))
            (c-statement (c-assign name "NULL"))
            (stub-parameter-conversion parameter)))))

;++ substrings
;++ byte subvectors

;;;; Stub Results

(define (stub-result/expression expression)
  (lambda (result-location)
    (c-statement (c-assign result-location expression))))

(define (stub-result/false) (stub-result/expression "S48_FALSE"))
(define (stub-result/true) (stub-result/expression "S48_TRUE"))
(define (stub-result/null) (stub-result/expression "S48_NULL"))
(define (stub-result/unspecific)
  (stub-result/expression "S48_UNSPECIFIC"))

(define (simple-stub-result expression type enter)
  (lambda (result-location)
    (c-statement
     (c-assign result-location
               (c-call enter (list (c-cast type expression)))))))

(define (stub-result/boolean expression)
  (simple-stub-result expression "int" "S48_ENTER_BOOLEAN"))

(define (stub-result/char expression)
  (simple-stub-result expression "char" "s48_enter_char"))

(define (stub-result/integral expression)
  (simple-stub-result expression "long" "s48_enter_integer"))

(define (stub-result/floating-point expression)
  (simple-stub-result expression "double" "s48_enter_double"))

(define (stub-result/string expression)
  (simple-stub-result expression "char *" "s48_enter_string"))

(define (stub-result/freed-string expression)
  (simple-stub-result expression "char *" "s48_enter_and_free_string"))

(define (stub-result-with-length expression length type enter)
  (lambda (result-location)
    (c-statement
     (c-assign result-location
               (c-call enter (list (c-cast type expression)
                                   (c-cast "long" length)))))))

(define (stub-result/substring expression length)
  (stub-result-with-length expression length
                           "char *" "s48_enter_substring"))

(define (stub-result/byte-vector expression length)
  (stub-result-with-length expression length
                           "char *" "s48_enter_byte_vector"))

(define (stub-result/freed-byte-vector expression length)
  (stub-result-with-length expression length
                           "char *" "s48_enter_and_free_byte_vector"))

(define (stub-result/fd->channel fd status id)
  (lambda (result-location)
    (c-statement
     (c-call "s48_enter_channel"
             (list (c-address-of result-location) fd status id)))))

;++ freed substrings
;++ freed byte subvectors

;;;;; Complex Result Patterns

;++ Sequences and conditionals should probably be moved elsewhere: they
;++ are not tied to Scheme48's C interface.

(define (stub-result/sequence item . items)
  (sequence-stub-result item items))

(define (sequence-stub-result item items)
  (let recur ((item item)
              (items items))
    (if (null? items)
        item
        (let ((item (if (string? item)
                        (c-line item)
                        item))
              (result (recur (car items) (cdr items))))
          (lambda (result-location)
            (output-sequence item (result result-location)))))))

(define (stub-result/conditional condition consequent alternative)
  (lambda (result-location)
    (c-if condition
          (c-block (consequent result-location))
          (c-block (alternative result-location)))))

(define (stub-result/cons car-result cdr-result)
  (lambda (result-location)
    (output-sequence
     (c-statement
      (c-assign result-location
                (c-call "s48_cons"
                        (list "S48_UNSPECIFIC" "S48_UNSPECIFIC"))))
     (car-result (c-call "S48_UNSAFE_CAR" (list result-location)))
     (cdr-result (c-call "S48_UNSAFE_CDR" (list result-location))))))

(define (stub-result/vector . element-results)
  (lambda (result-location)
    (output-sequence
     (c-statement
      (c-assign result-location
                (c-call "s48_make_vector"
                        (list (length element-results)
                              "S48_UNSPECIFIC"))))
     (output-list
      (loop ((FOR element-result (IN-LIST element-results))
             (LET index 0 (+ index 1))
             (FOR assignments
                  (LISTING (element-result
                            (c-call "S48_UNSAFE_VECTOR_REF"
                                    (list result-location index))))))
            => assignments)))))

(define (stub-result/syscall-error syscall error-code)
  syscall                               ;ignore
  (lambda (result-location)
    result-location                     ;ignore
    (c-statement (c-call "s48_raise_os_error" (list error-code)))))

(define (stub-result/out-of-memory-error)
   (lambda (result-location)
     result-location                    ;ignore
     (c-statement (c-call "s48_raise_out_of_memory_error" '()))))

(define (s48-value-type)
  "s48_value")

(define (s48-export-binding binding-name expression)
  (c-statement
   (c-call "s48_define_exported_binding"
           (list binding-name expression))))

(define (s48-import-binding binding-name location)
  (c-statement
   (c-assign location
             (c-call "s48_get_imported_binding"
                     (list binding-name)))))

(define (s48-declare-false variable)
  (c-declare (s48-value-type)
             (c-initialize variable "S48_FALSE")))

(define (s48-declare-unspecific variable)
  (c-declare (s48-value-type)
             (c-initialize variable "S48_UNSPECIFIC")))

(define (s48-gc-protect-global location)
  (c-statement (c-call "S48_GC_PROTECT_GLOBAL" (list location))))

(define (s48-declare-gc-protect n)
  (c-statement (c-call "S48_DECLARE_GC_PROTECT" (list n))))

(define (s48-gc-protect locations)
  (c-statement
   (c-call (output-sequence "S48_GC_PROTECT_" (length locations))
           locations)))

(define (s48-gc-unprotect)
  (c-statement (c-call "S48_GC_UNPROTECT" '())))
