;;; -*- Mode: Scheme; scheme48-package: usual-stubber -*-

;;;; Scheme48 Stubber
;;;; Usual Stub Environment

;;; This code is written by Taylor R. Campbell and placed in the Public
;;; Domain.  All warranties are disclaimed.

(define (usual-stub-file input-pathname c-pathname scheme-pathname)
  (stub-file input-pathname
             c-pathname
             scheme-pathname
             (make-usual-stub-environment)))

(define (make-usual-stub-environment)
  (copy-stub-environment usual-stub-environment))

(define usual-stub-environment
  (make-stub-environment
   (output-sequence
    (c-line (c-comment "This file is automatically generated.   "))
    (c-line (c-comment "Modify at your own peril!               "))
    newline
    (cpp-include "s48-stubber.h"))
   '()                                  ;Scheme prelude
   p                                    ;Scheme writer (pretty-print)
   symbol->string                       ;Binding namer
   (lambda (scheme-name) (usual-c-stub-name scheme-name))
   (lambda (scheme-name) (usual-c-variable-name scheme-name))
   '()                                  ;Form handlers
   '()                                  ;Parameter handlers
   '()                                  ;Expression handlers
   ))

(define (usual-c-stub-name scheme-name)
  (output-sequence "s48_stub__" (usual-c-variable-name scheme-name)))

(define (usual-c-variable-name scheme-name)       ;++ Silliness
  (output-list
   (map (lambda (char)
          (cond ((char=? char #\-) #\_)
                ((char=? char #\?) #\P)
                ((or (char-alphabetic? char)
                     (char-numeric? char))
                 char)
                (else
                 (output-sequence "_0x"
                                  (number->string (char->ascii char)
                                                  16)
                                  "_"))))
        (string->list (symbol->string scheme-name)))))

(define-form-handler INCLUDE usual-stub-environment
    (form environment)
  (('INCLUDE (? string? headers) ___)
   (stub/c-declarations (output-list (map cpp-include headers)))))

(define-form-handler SYSTEM-INCLUDE usual-stub-environment
    (form environment)
  (('SYSTEM-INCLUDE (? string? headers) ___)
   (stub/c-declarations
    (output-list (map cpp-system-include headers)))))

(define-form-handler DEFINE usual-stub-environment
    (form environment)
  (('DEFINE (? symbol? scheme-name) expression)
   (binding-definition scheme-name expression environment))
  (('DEFINE ((? symbol? scheme-name) parameter-specifiers ___)
     ((? string? c-declarations) ___)
     expression)
   (lambda-definition scheme-name
                      parameter-specifiers
                      #f
                      c-declarations
                      (list expression)
                      environment))
  (('DEFINE ((? symbol? scheme-name) parameter-specifiers ___)
     ((? string? c-declarations) ___)
     expression
     scheme-output-processor)
   (lambda-definition scheme-name
                      parameter-specifiers
                      scheme-output-processor
                      c-declarations
                      (list expression)
                      environment)))

(define (binding-definition scheme-name expression environment)
  (stub/binding scheme-name
                (symbol->binding-name scheme-name environment)
                expression
                environment))

(define (lambda-definition scheme-name
                           parameter-specifiers
                           scheme-output-processor
                           c-declarations
                           body
                           environment)
  (let ((binding-name (symbol->binding-name scheme-name environment))
        (c-name (symbol->c-stub-name scheme-name environment)))
    (receive (scheme-names scheme-argument-expressions c-parameters)
             (process-parameter-specifiers parameter-specifiers
                                           environment)
      (stub/function scheme-name
                     binding-name
                     c-name
                     scheme-names
                     scheme-argument-expressions
                     scheme-output-processor
                     c-parameters
                     c-declarations
                     (process-body body environment)))))

(define (process-parameter-specifiers parameter-specifiers
                                      environment)
  ;++ It would be nice to be able to accumulate multiple lists with
  ;++ multiple values here.
  (loop continue ((FOR specifier (IN-LIST parameter-specifiers))
                  (LET scheme-names '())
                  (LET scheme-argument-expressions '())
                  (LET c-parameters '()))
    => (values (reverse scheme-names)
               (reverse scheme-argument-expressions)
               (reverse c-parameters))
    (receive (scheme-name c-name form)
             (parse-parameter-specifier specifier environment)
      (receive (scheme-argument-expression c-parameter)
               ((process-parameter form environment) scheme-name
                                                     c-name)
        (continue (=> scheme-names (cons scheme-name scheme-names))
                  (=> scheme-argument-expressions
                      (cons scheme-argument-expression
                            scheme-argument-expressions))
                  (=> c-parameters
                      (cons c-parameter c-parameters)))))))

(define (parse-parameter-specifier specifier environment)
  (match specifier
    (((? symbol? scheme-name) form)
     (values scheme-name
             (symbol->c-variable-name scheme-name environment)
             form))
    ((((? symbol? scheme-name) (? string? c-name)) form)
     (values scheme-name c-name form))
    (else
     (error "Malformed parameter specifier:" specifier))))

(define (process-body body environment)
  (let recur ((body body))
    (let ((item (car body))
          (body (cdr body)))
      (if (null? body)
          (process-expression item environment)
          (stub-result/sequence item (recur body))))))

(define-form-handler DEFINE-VOID-SYSCALL usual-stub-environment
    (form environment)
  (('DEFINE-VOID-SYSCALL ((? symbol? name) parameter-specifiers ___)
     ((? string? declarations) ___)
     (? string? syscall) ((? string? arguments) ___)
     result)
   (syscall-definition name parameter-specifiers #f
                       "int"
                       (generate-c-temporary "status" environment)
                       declarations syscall arguments result
                       environment))
  (('DEFINE-VOID-SYSCALL ((? symbol? name) parameter-specifiers ___)
     ((? string? declarations) ___)
     (? string? syscall) ((? string? arguments) ___)
     result
     output)
   (syscall-definition name parameter-specifiers output
                       "int"
                       (generate-c-temporary "status" environment)
                       declarations syscall arguments result
                       environment)))

(define-form-handler DEFINE-UINT-SYSCALL usual-stub-environment
    (form environment)
  (('DEFINE-UINT-SYSCALL ((? symbol? name) parameter-specifiers ___)
     ((? string? declarations) ___)
     (? string? value-variable)
     (? string? syscall) ((? string? arguments) ___)
     result)
   (syscall-definition name parameter-specifiers #f
                       "int" value-variable
                       declarations syscall arguments result
                       environment))
  (('DEFINE-UINT-SYSCALL ((? symbol? name) parameter-specifiers ___)
     ((? string? declarations) ___)
     (? string? value-variable)
     (? string? syscall) ((? string? arguments) ___)
     result
     output)
   (syscall-definition name parameter-specifiers output
                       "int" value-variable
                       declarations syscall arguments result
                       environment)))

(define (syscall-definition name parameter-specifiers output
                            value-type value-variable
                            declarations syscall arguments result
                            environment)
  (lambda-definition
   name parameter-specifiers output
   (cons (output-sequence value-type " " value-variable ";")
         declarations)
   (list (c-statement
          (c-call "S48_SYSCALL_UINT"
                  (list (output-sequence "syscall_" syscall)
                        value-variable
                        (c-call syscall arguments))))
         result)
   environment))

(define-form-handler DEFINE-CONSTANT-ENUMERATION
    usual-stub-environment
    (form environment)
  (('DEFINE-CONSTANT-ENUMERATION name rtd type-predicate
     instance-vector
     name-accessor
     index-accessor
     available-predicate
     (AND c-cruft ((? string? c-type)
                   (? string? c-mapper)
                   (? string? c-unmapper)))
     (((? symbol? enumerand-names) enumerand-expressions) ___))
   (*define-enumeration-conversions name index-accessor
                                    c-type c-mapper c-unmapper
                                    environment)
   ;** This must be synchronized with the definition in codegen.scm!
   ;++ ugh...
   (stub/enumeration
    name rtd
    type-predicate
    instance-vector
    (symbol->binding-name instance-vector environment)
    name-accessor
    index-accessor
    available-predicate
    (symbol->binding-name available-predicate environment)
    enumerand-names
    enumerand-expressions
    environment
    c-type c-mapper c-unmapper)))

(define (*define-enumeration-conversions name index-accessor
                                         c-type c-mapper c-unmapper
                                         environment)
  (*define-parameter-handler
   name environment
   (enumerand-mapper name c-type c-mapper index-accessor))
  (*define-expression-handler
   name environment
   (enumerand-unmapper name c-type c-unmapper)))

(define (enumerand-mapper name c-type c-mapper index-accessor)
  (lambda (form environment)
    (if (not (eq? form name))
        (error "Malformed parameter:" form)
        (lambda (scheme-name c-name)
          (values `(,index-accessor ,scheme-name)
                  (simple-stub-parameter c-name c-type c-mapper))))))

(define (enumerand-unmapper name c-type c-unmapper)
  (lambda (form environment)
    (match form
      (((? (lambda (x) (eq? x name))) expression)
       (stub-result/expression
        (c-call c-unmapper (list expression)))))))

(define-simple-parameter-handler SCHEME-DATUM usual-stub-environment
  stub-parameter/scheme)

(define-simple-parameter-handler BOOLEAN usual-stub-environment
  stub-parameter/boolean)

(define-simple-parameter-handler CHAR usual-stub-environment
  stub-parameter/char)

(define-simple-parameter-handler CHANNEL->FD usual-stub-environment
  stub-parameter/channel->fd)

(define-parameter-handler INTEGRAL usual-stub-environment
    (form environment)
  (('INTEGRAL (? string? type))
   (lambda (scheme-name c-name)
     (values `(IF (AND (INTEGER? ,scheme-name) (EXACT? ,scheme-name))
                  ,scheme-name
                  (ERROR "Non-integer parameter:"
                         ,scheme-name
                         ',scheme-name))
             (stub-parameter/integral type c-name)))))

(define-parameter-handler FLOATING-POINT usual-stub-environment
    (form environment)
  (('FLOATING-POINT (? string? type))
   (lambda (scheme-name c-name)
     (values `(IF (REAL? ,scheme-name)
                  (EXACT->INEXACT ,scheme-name)
                  (ERROR "Non-real parameter:"
                         ,scheme-name
                         ',scheme-name))
             (stub-parameter/floating-point type c-name)))))

(define-simple-parameter-handler SHARED-STRING
  usual-stub-environment
  stub-parameter/shared-string)

(define-simple-parameter-handler COPIED-STRING
  usual-stub-environment
  stub-parameter/copied-string)

(define-parameter-handler SHARED-BYTE-VECTOR usual-stub-environment
    (form environment)
  (('SHARED-BYTE-VECTOR (? string? type) (? string? length-name))
   (lambda (scheme-name c-name)
     (*byte-vector-parameter stub-parameter/shared-byte-vector
                             type length-name scheme-name c-name))))

(define-parameter-handler COPIED-BYTE-VECTOR usual-stub-environment
    (form environment)
  (('COPIED-BYTE-VECTOR (? string? type) (? string? length-name))
   (lambda (scheme-name c-name)
     (*byte-vector-parameter stub-parameter/copied-byte-vector
                             type length-name scheme-name c-name))))

(define (*byte-vector-parameter stub-parameter/byte-vector
                                type length-name scheme-name c-name)
  (values `(IF (BYTE-VECTOR? ,scheme-name)
               ,scheme-name
               (ERROR "Non-byte-vector parameter:"
                      ,scheme-name
                      ',scheme-name))
          (stub-parameter/byte-vector type c-name length-name)))

(define-parameter-handler FALSE->NULL usual-stub-environment
    (form environment)
  (('FALSE->NULL parameter)
   (let ((processor (process-parameter parameter environment)))
     (lambda (scheme-name c-name)
       (receive (scheme-conversion-form parameter)
                (processor scheme-name c-name)
         (values `(AND ,scheme-name ,scheme-conversion-form)
                 (stub-parameter/false->null c-name parameter)))))))

(define-simple-expression-handler SCHEME-DATUM
  usual-stub-environment
  stub-result/expression)

(define-constant-expression-handler TRUE usual-stub-environment
  stub-result/true)

(define-constant-expression-handler FALSE usual-stub-environment
  stub-result/false)

(define-constant-expression-handler NULL usual-stub-environment
  stub-result/null)

(define-constant-expression-handler UNSPECIFIC
  usual-stub-environment
  stub-result/unspecific)

(define-simple-expression-handler BOOLEAN usual-stub-environment
  stub-result/boolean)

(define-simple-expression-handler CHAR usual-stub-environment
  stub-result/char)

(define-simple-expression-handler INTEGRAL usual-stub-environment
  stub-result/integral)

(define-simple-expression-handler FLOATING-POINT
  usual-stub-environment
  stub-result/floating-point)

(define-simple-expression-handler STRING usual-stub-environment
  stub-result/string)

(define-simple-expression-handler FREED-STRING
  usual-stub-environment
  stub-result/freed-string)

(define-expression-handler SUBSTRING usual-stub-environment
    (form environment)
  (('SUBSTRING string-expression length-expression)
   (stub-result/substring string-expression length-expression)))

(define-expression-handler BYTE-VECTOR usual-stub-environment
    (form environment)
  (('BYTE-VECTOR pointer-expression length-expression)
   (stub-result/byte-vector pointer-expression length-expression)))

(define-expression-handler FREED-BYTE-VECTOR usual-stub-environment
    (form environment)
  (('FREED-BYTE-VECTOR pointer-expression length-expression)
   (stub-result/freed-byte-vector pointer-expression
                                  length-expression)))

;;;; Complex Expression Combinators

(define-expression-handler BEGIN usual-stub-environment
    (form environment)
  (('BEGIN item items ___)
   (process-body (cons item items) environment)))

(define-expression-handler IF usual-stub-environment
    (form environment)
  (('IF condition consequent alternative)
   (stub-result/conditional
    (process-condition condition environment)
    (process-expression consequent environment)
    (process-expression alternative environment))))

(define (process-condition condition environment)    ;++ ad-hockery
  (let process ((condition condition))
    (match condition
      ((? string?)          condition)
      (('NOT negated)       (c-not (process negated)))
      (('AND conjuncts ___) (c-and* (map process conjuncts)))
      (('OR  disjuncts ___) (c-or*  (map process disjuncts)))
      (else (error "Malformed condition:" condition)))))

(define-expression-handler COND usual-stub-environment
    (form environment)
  (('COND (AND clauses (conditions bodies ___)) ___)
   (if (null? clauses)
       (error "Empty COND.")
       (let recur ((clauses clauses))
         (match clauses
           ((('ELSE body ___))
            (process-body body environment))
           ((('ELSE body ___) . clauses)
            (error "Trailing clauses after ELSE in COND:"
                   clauses
                   `(ELSE ,@body)))
           (((condition body ___) . clauses)
            (stub-result/conditional
             (process-condition condition environment)
             (process-body body environment)
             (recur clauses))))))))

(define-expression-handler CONS usual-stub-environment
    (form environment)
  (('CONS car-expression cdr-expression)
   (stub-result/cons
    (process-expression car-expression environment)
    (process-expression cdr-expression environment))))

(define-expression-handler VECTOR usual-stub-environment
    (form environment)
  (('VECTOR element-expressions ___)
   (apply stub-result/vector
          (map (lambda (element-expression)
                 (process-expression element-expression environment))
               element-expressions))))

(define-expression-handler FD->CHANNEL usual-stub-environment
    (form environment)
  (('FD->CHANNEL fd-expression status-expression id-expression)
   (stub-result/fd->channel fd-expression
                            status-expression
                            id-expression)))

;;;; Error Results

(define-expression-handler SYSCALL-ERROR usual-stub-environment
    (form environment)
  (('SYSCALL-ERROR)
   (stub-result/syscall-error #f "errno"))
  (('SYSCALL-ERROR syscall)
   (stub-result/syscall-error syscall "errno"))
  (('SYSCALL-ERROR syscall error-code)
   (stub-result/syscall-error syscall error-code)))

(define-expression-handler OUT-OF-MEMORY-ERROR
    usual-stub-environment
    (form environment)
  (('OUT-OF-MEMORY-ERROR)
   (stub-result/out-of-memory-error)))
